/*
    AQTRONIX C++ Library
    Copyright 2005-2006 Parcifal Aertssen

    This file is part of AQTRONIX C++ Library.

    AQTRONIX C++ Library is free software; you can redistribute it
	and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

    AQTRONIX C++ Library is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AQTRONIX C++ Library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
    MA  02111-1307  USA
*/
// HTTPFirewall.cpp: implementation of the CHTTPFirewall class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HTTPFirewall.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHTTPFirewall::CHTTPFirewall(CHTTPFirewallSettings& _Settings, CLogger& _logger): Settings(_Settings),CFirewall(_Settings,_logger)
{

}

CHTTPFirewall::~CHTTPFirewall()
{

}

Action CHTTPFirewall::ScanRawHeaders(CString &Headers)
{
	Action action;
	CString strhLog = Headers.Left(4096); //for log file

	//HEADERS: generic header length
	if(Settings.Headers.MaxLength.Action && Settings.Headers.MaxLength.Value>0){
		if(ScanVariableLength(Headers,Settings.Headers.MaxLength.Value,CR,LF)){
			action.Set(Settings.Headers.MaxLength.Action);
			Alert("Maximum generic header length exceeded",Settings.Headers.MaxLength.Action);
		}
	}
		
	//HEADERS: check for encoding abuse
	if(Settings.Headers.EncodingExploits){
		if(ScanHexEncodingAbuse(Headers,Settings.EncodingExploits.Keywords)){
			action.Set(Settings.Headers.EncodingExploits);
			Alert("Encoding exploit in headers",Settings.Headers.EncodingExploits);
		}
	}
	
	Headers = CURL::Decode(Headers);

	//HEADERS: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
	if(Settings.Headers.EncodingExploits && Settings.EncodingExploits.ScanDoubleEncoding){
		if(CanHexDecode(Headers,false)){
			action.Set(Settings.Headers.EncodingExploits);
			Alert("Embedded encoding exploit in headers",Settings.Headers.EncodingExploits);
		}
	}

	//HEADERS: Deny directory traversal (parent path)
	if(Settings.Headers.DirectoryTraversal){
		if(ScanDirectoryTraversal(Headers)){
			action.Set(Settings.Headers.DirectoryTraversal);
			Alert("Directory traversal not allowed in headers",Settings.Headers.DirectoryTraversal);
		}
	}

	//HEADERS: Deny high bit characters (shellcode)
	if(Settings.Headers.HighBitShellcode){
		if(ScanHighBitShellCode(Headers)){
			action.Set(Settings.Headers.HighBitShellcode);
			Alert("High bit characters (shellcode) not allowed in headers",Settings.Headers.HighBitShellcode);
		}
	}

	CStringHelper::Safe_MakeLower(Headers);

	//HEADERS: Deny SQL injection (words in list should be lowercase)
	if(Settings.Headers.SQLInjection){	
		int sqlcount(Settings.SQLi.Keywords.Find(";")!=NULL?-1:0);
		if(Settings.SQLi.Keywords.Find("/*")!=NULL)	//Accept: */*
			sqlcount--;
		if(Settings.SQLi.Keywords.Find("*/")!=NULL)
			sqlcount--;
		CStringList Matches;
		sqlcount += ScanSQLInjection(Headers,Settings.SQLi.Keywords,Matches);
		if(sqlcount>0){
			if(sqlcount>Settings.SQLi.AllowedCount){
				action.Set(Settings.Headers.SQLInjection);
				Alert("Possible SQL injection in headers: " + CStringHelper::ToString(Matches,","),Settings.Headers.SQLInjection);
			}else{
				logger.Append("WARNING: SQL keyword found in headers (" + CStringHelper::Safe_IntToAscii(Settings.SQLi.AllowedCount - sqlcount + 1) + " more will block request): " + CStringHelper::ToString(Matches,","));
			}
		}
	}

	//HEADERS: Deny certain sequences (words in list should be lowercase)
	if(Settings.Headers.DenySequences.Action){
		if(DenySequences(Headers,Settings.Headers.DenySequences.List,"'","' not allowed in headers",Settings.Headers.DenySequences.Action)){
			action.Set(Settings.Headers.DenySequences.Action);
		}
	}

	//HEADERS: Deny regex
	if(Settings.Headers.DenyRegex.Action){
		if(DenyRegex(Headers,Settings.Headers.DenyRegex.Map,"'","' regex not allowed in headers",Settings.Headers.DenyRegex.Action)){
			action.Set(Settings.Headers.DenyRegex.Action);
		}
	}

	if (!action.IsEmpty()){
		strhLog.Replace('\r',' ');
		strhLog.Replace('\n',' ');
		logger.Append(strhLog);
	}

	return action;
}

Action CHTTPFirewall::ScanRawData(CString &Data)
{
	//return ScanEntity(Data,CString("multipart/form-data; boundary=---------------------------")); //for testing on IIS 5

	return ScanEntity(Data,CString("application/x-www-form-urlencoded"));
}

Action CHTTPFirewall::ScanEntity(CString &Data, CString ContentType, unsigned int depth)
{
	Action action;
	//logger.Append(Data);
	CStringHelper::Safe_MakeLower(ContentType);
	//TODO: exploits?
	/* 
		- recursive
		- text/plain
	 */

	if(ContentType.Left(19)=="multipart/form-data" && depth<5){
		//multiple entities
		depth++;
		int posstart = ContentType.Find("boundary=");
		if(posstart>-1){
			int posend = ContentType.Find(CRLF,posstart);
			if(posend==-1)
				posend = ContentType.GetLength()-1;
			CString boundary = "--" + ContentType.Mid(posstart+9,posend-posstart-8); //RFC: "--" followed by boundary parameter
			//logger.Append(boundary);

			CStringList entities;
			CStringHelper::Split(entities, Data, boundary);

			//recursive
			POSITION pos = entities.GetHeadPosition();
			while(pos!=NULL){
				CString entity = entities.GetNext(pos);

				//content-type
				CString ct = "text/plain"; //default according to RFC
				posstart = entity.Find("Content-Type: "); //length: 14
				if(posstart>-1 && posstart<entity.GetLength()-15){
					posend = entity.Find(CRLF,posstart+14);
					if(posend>-1)
						ct = entity.Mid(posstart+14,posend-posstart-14);
				}

				//data
				posstart = entity.Find(CRLFCRLF);
				if(posstart>0 && posstart<entity.GetLength()-4)
					entity = entity.Mid(posstart+4);

				//logger.Append(ct);
				action |= ScanEntity(entity,ct,depth);
			}
		}else{
			//no boundary?
			action |= ScanEntity(Data,CString("application/x-www-form-urlencoded"));
		}
	}else{
		//single entity
		CString strdLog = Data.Left(1024); //for log file
		bool isurlencoded = (ContentType=="" || ContentType.Left(33)=="application/x-www-form-urlencoded"); //assume empty content-type url encoded as well to scan exploits (RFC: empty = text/plain)
		bool istextplain =	(ContentType=="" || ContentType.Left(10)=="text/plain");
		bool isbinary =		(!isurlencoded && ContentType.Left(5)!="text/");

		//DATA: variable size
		if(Settings.Global.MaxPostVariableLength.Action && Settings.Global.MaxPostVariableLength.Value>0 && Data.Find('=')!=-1){
			if(	( isurlencoded && ScanVariableLength(Data,Settings.Global.MaxPostVariableLength.Value,'&','&') ) ||
				( istextplain && ScanVariableLength(Data,Settings.Global.MaxPostVariableLength.Value,CR,LF) )
			){
				action.Set(Settings.Global.MaxPostVariableLength.Action);
				Alert("Maximum variable length exceeded in data",Settings.Global.MaxPostVariableLength.Action);
			}
		}

		if(isurlencoded){

			//DATA: check for encoding abuse
			if(Settings.Global.PostEncodingExploits){
				if(ScanHexEncodingAbuse(Data,Settings.EncodingExploits.Keywords)){
					action.Set(Settings.Global.PostEncodingExploits);
					Alert("Encoding exploit in data",Settings.Global.PostEncodingExploits);
				}
			}

			Data = CURL::Decode(Data);

			//DATA: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
			if(Settings.Global.PostEncodingExploits && Settings.EncodingExploits.ScanDoubleEncoding){
				if(CanHexDecode(Data,false)){
					action.Set(Settings.Global.PostEncodingExploits);
					Alert("Encoding exploit in data (embedded encoding)",Settings.Global.PostEncodingExploits);
				}		
			}
		}

		if(!isbinary){

			//DATA: Deny directory traversal (parent path)
			if(Settings.Global.PostDirectoryTraversal){
				if(ScanDirectoryTraversal(Data)){
					action.Set(Settings.Global.PostDirectoryTraversal);
					Alert("Directory traversal not allowed in data",Settings.Global.PostDirectoryTraversal);
				}
			}

			//DATA: Deny high bit characters (shellcode)
			if(Settings.Global.PostHighBitShellcode){
				if(ScanHighBitShellCode(Data)){
					action.Set(Settings.Global.PostHighBitShellcode);
					Alert("High bit characters (shellcode) not allowed in data",Settings.Global.PostHighBitShellcode);
				}
			}

		}

		CStringHelper::Safe_MakeLower(Data);

		if(!isbinary){

			//DATA: Deny SQL injection (words in list should be lowercase)
			if(Settings.Global.PostSQLInjection){
				int sqlcount(0);
				CStringList Matches;
				sqlcount += ScanSQLInjection(Data,Settings.SQLi.Keywords, Matches);
				if(sqlcount>0){
					if(sqlcount>Settings.SQLi.AllowedCount){
						action.Set(Settings.Global.PostSQLInjection);
						Alert("Possible SQL injection in data: " + CStringHelper::ToString(Matches,","),Settings.Global.PostSQLInjection);
					}else{
						logger.Append("WARNING: SQL keyword found in data (" + CStringHelper::Safe_IntToAscii(Settings.SQLi.AllowedCount - sqlcount + 1) + " more will block request): " + CStringHelper::ToString(Matches,","));
					}
				}
			}

		}

		//DATA: deny certain sequences (list should be lowercase)
		if(Settings.Global.DenyPostSequences.Action){
			if(DenySequences(Data,Settings.Global.DenyPostSequences.List,"'","' not allowed in data",Settings.Global.DenyPostSequences.Action)){
				action.Set(Settings.Global.DenyPostSequences.Action);
			}
		}

		//DATA: deny regex
		if(Settings.Global.DenyPostRegex.Action){
			if(DenyRegex(Data,Settings.Global.DenyPostRegex.Map,"'","' regex not allowed in data",Settings.Global.DenyPostRegex.Action)){
				action.Set(Settings.Global.DenyPostRegex.Action);
			}
		}

		if(!action.IsEmpty()){
			strdLog.Replace('\r',' ');
			strdLog.Replace('\n',' ');
			logger.Append(strdLog);
		}

	}
	return action;
}

Action CHTTPFirewall::ScanAccount(CString &User)
{
	Action action;
	if(User==""){
		return action;
	}else{
		CString USER(User);
		CStringHelper::Safe_MakeUpper(USER);

		//AUTH - block logon attempts for certain critical accounts
		if(Settings.Authentication.SystemAccounts){
			CString UserLeft = USER.Left(5);
			if(	UserLeft=="IUSR_" ||	//InternetUser
				UserLeft=="IWAM_" ||	//InternetWAM
				UserLeft=="VUSR_" ||	//Visual studio server components
				USER=="SYSTEM" ||		//SYSTEM account
				USER=="KRBTGT" ||		//Kerberos
				USER=="ASPNET" ||		//ASP.NET (out of process)
				USER=="ASPNETUSER" ||	//ASP.NET (out of process) Windows 2003
				USER=="TSINTERNETUSER" ||//Terminal services account (internet account)
				USER=="NETWORK SERVICE"){
					action.Set(Settings.Authentication.SystemAccounts);
					Alert("Attempt to authenticate with system critical account",Settings.Authentication.SystemAccounts);
			}
		}

		//AUTH - only allow certain accounts
		if(Settings.Authentication.AllowAccounts.Action){
			if(!IsInListCompareNoCase(User,Settings.Authentication.AllowAccounts.List)){
				action.Set(Settings.Authentication.AllowAccounts.Action);
				Alert("Account not in allowed list",Settings.Authentication.AllowAccounts.Action);
			}
		}

		//AUTH - Deny certain accounts
		if(Settings.Authentication.DenyAccounts.Action){
			if(IsInListCompareNoCase(User,Settings.Authentication.DenyAccounts.List)){
				action.Set(Settings.Authentication.DenyAccounts.Action);
				Alert("Account not allowed to authenticate",Settings.Authentication.DenyAccounts.Action);
			}
		}
		return action;
	}
}

Action CHTTPFirewall::ScanFilename(CFileName &fn)
{
	Action action;
	char c;

	//FILENAME: Deny default document
	if(Settings.Filename.DefaultDocument){
		if(fn.FileName == ""){
			action.Set(Settings.Filename.DefaultDocument);
			Alert("Default document not allowed",Settings.Filename.DefaultDocument);
		}
	}

	//FILENAME: Deny certain characters in the filename (characters are case sensitive!)
	if(Settings.Filename.Characters.Action){
		for(int i=0;i<Settings.Filename.Characters.Value.GetLength();i++){
			c = Settings.Filename.Characters.Value.GetAt(i);
			if(DenyChar(fn.FileName,c,"'" + CString(c) + "' not allowed in filename",Settings.Filename.Characters.Action))
				action.Set(Settings.Filename.Characters.Action);
		}
	}

	//FILENAME: Deny certain cgi applications (applications in list should be lowercase).
	if(Settings.Filename.DenyFiles.Action){
		if(DenySequences(fn.FileNameLCase,Settings.Filename.DenyFiles.List,"accessing/running '","' file",Settings.Filename.DenyFiles.Action)){
			action.Set(Settings.Filename.DenyFiles.Action);
		}
	}

	//EXTENSION: Only allow certain extensions (extensions in the list should be lowercase)
	if(Settings.Extensions.AllowExtensions.Action){
		if(!IsInList(fn.ExtensionLCase,Settings.Extensions.AllowExtensions.List)){
			action.Set(Settings.Extensions.AllowExtensions.Action);
			Alert("accessing/running '" + fn.Extension + "' file",Settings.Extensions.AllowExtensions.Action);
		}
	}

	//EXTENSION: Deny certain extensions (extensions in the list should be lowercase)
	if(Settings.Extensions.DenyExtensions.Action){
		if(DenySequences(fn.ExtensionLCase,Settings.Extensions.DenyExtensions.List,"accessing/running '","' file",Settings.Extensions.DenyExtensions.Action)){
			action.Set(Settings.Extensions.DenyExtensions.Action);
		}
	}

	return action;
}

Action CHTTPFirewall::ScanURL(const char *p)
{
	Action action;
	CString URL(p);
	CStringHelper::Safe_MakeLower(URL);

	//URL: Only allow certain characters as the start of a URL (list should be lowercase)
	if(Settings.URL.AllowURL.Action){
		if(!IsStartInList(URL,Settings.URL.AllowURL.List)){
			action.Set(Settings.URL.AllowURL.Action);
			Alert("URL not in allowed list",Settings.URL.AllowURL.Action);
		}
	}

	//URL: Deny parent path
	if (Settings.URL.ParentPath){
		if(DenyString(p,"..","'..' not allowed in URL (parent path)",Settings.URL.ParentPath)){
			action.Set(Settings.URL.ParentPath);
		}
	}

	//URL: Deny trailing dot in directory
	if (Settings.URL.TrailingDotInDir){
		if(DenyString(p,"./","'./' not allowed in URL (trailing dot on a directory name)",Settings.URL.TrailingDotInDir)){
			action.Set(Settings.URL.TrailingDotInDir);
		}
	}
	
	//URL: Deny backslash
	if (Settings.URL.Backslash){
		if(DenyChar(p,'\\',"'\\' not allowed in URL (backslashes)",Settings.URL.Backslash)){
			action.Set(Settings.URL.Backslash);
		}
	}

	//URL: Deny alternate stream access
	if (Settings.URL.AlternateStream){
		if(DenyChar(p,':',"':' not allowed in URL (alternate stream access)",Settings.URL.AlternateStream)){
			action.Set(Settings.URL.AlternateStream);
		}
	}

	//URL: Deny escaping
	if (Settings.URL.Escaping){
		if(DenyChar(p,'%',"'%' not allowed in URL (hex encoding)",Settings.URL.Escaping)){
			action.Set(Settings.URL.Escaping);
		}
	}

	//URL: Deny running multiple applications
	if (Settings.URL.MultipleCGI){
		if(DenyChar(p,'&',"'&' not allowed in URL (running multiple CGI processes in single request)",Settings.URL.MultipleCGI)){
			action.Set(Settings.URL.MultipleCGI);
		}
	}
	
	//URL: Deny certain characters
	if(Settings.URL.Characters.Action){
		char c;
		for(int i=0;i<Settings.URL.Characters.Value.GetLength();i++){
			c = Settings.URL.Characters.Value.GetAt(i);
			if(DenyChar(p,c,"'" + CString(c) + "' not allowed in URL",Settings.URL.Characters.Action))
				action.Set(Settings.URL.Characters.Action);
		}
	}

	//URL: Deny high bit characters (shellcode)
	if(Settings.URL.HighBitShellcode){
		if(ScanHighBitShellCode(p)){
			action.Set(Settings.URL.HighBitShellcode);
			Alert("High bit characters (shellcode) not allowed in URL",Settings.URL.HighBitShellcode);
		}
	}

	//URL: Deny special whitespace characters
	if(Settings.URL.SpecialWhitespace){
		if(DenySpecialWhitespace(p, " not allowed in URL",Settings.URL.SpecialWhitespace)){
			action.Set(Settings.URL.SpecialWhitespace);
		}
	}

	//URL: Deny certain sequences (list should be lowercase)
	if(Settings.URL.DenySequences.Action){
		if(DenySequences(URL,Settings.URL.DenySequences.List,"'","' not allowed in URL",Settings.URL.DenySequences.Action)){
			action.Set(Settings.URL.DenySequences.Action);
		}
	}

	//URL: deny regex
	if(Settings.URL.DenyRegex.Action){
		if(DenyRegex(URL,Settings.URL.DenyRegex.Map,"'","' regex not allowed in URL",Settings.URL.DenyRegex.Action)){
			action.Set(Settings.URL.DenyRegex.Action);
		}
	}
	
	return action;
}

Action CHTTPFirewall::ScanURLDoS(CString& Url)
{

	Action action;

	//URL - Request Limits
	if(Settings.URL.RequestsLimit.Action && Url!="/"){
		//add request to cache
		Statistics.URL_RequestsLimit.Add(Url);

		//block if number of requests exceeds certain treshold
		if(Statistics.URL_RequestsLimit.Count(Url)>Settings.URL.RequestsLimit.MaxCount){
			action.Set(Settings.URL.RequestsLimit.Action);
			Alert("Reached URL requests limit",Settings.URL.RequestsLimit.Action);
		}
	}

	return action;
}

Action CHTTPFirewall::ScanRawQuerystring(CString& QueryString)
{
	Action action;
	int l(0);
	if((l = QueryString.GetLength())==0){
		return action;
	}else{
		logger.Append(QueryString.Left(9000));

		//QUERYSTRING: size
		if(Settings.QueryString.MaxLength.Action){
			if (l>Settings.QueryString.MaxLength.Value){
				action.Set(Settings.QueryString.MaxLength.Action);
				Alert("QueryString is too long!",Settings.QueryString.MaxLength.Action);
			}
		}

		//QUERYSTRING: variable size
		if(Settings.QueryString.MaxVariableLength.Action && Settings.QueryString.MaxVariableLength.Value>0 && QueryString.Find('=')!=-1){
			if(ScanVariableLength(QueryString,Settings.QueryString.MaxVariableLength.Value,'&','&')){
				action.Set(Settings.QueryString.MaxVariableLength.Action);
				Alert("Maximum variable length exceeded in querystring",Settings.QueryString.MaxVariableLength.Action);
			}
		}

		//QUERYSTRING: encoding abuse
		if(Settings.QueryString.EncodingExploits){
			if(ScanHexEncodingAbuse(QueryString,Settings.EncodingExploits.Keywords)){
				action.Set(Settings.QueryString.EncodingExploits);
				Alert("Encoding exploit in querystring",Settings.QueryString.EncodingExploits);
			}
		}
		
		QueryString = CURL::Decode(QueryString);

		//QUERYSTRING: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
		if(Settings.QueryString.EncodingExploits && Settings.EncodingExploits.ScanDoubleEncoding){
			if(CanHexDecode(QueryString,false)){
				action.Set(Settings.QueryString.EncodingExploits);
				Alert("Querystring embedded encoding exploit",Settings.QueryString.EncodingExploits);
			}
		}

		//QUERYSTRING: Deny directory traversal (parent path)
		if(Settings.QueryString.DirectoryTraversal){
			if(ScanDirectoryTraversal(QueryString)){
				action.Set(Settings.QueryString.DirectoryTraversal);
				Alert("Directory traversal not allowed in querystring",Settings.QueryString.DirectoryTraversal);
			}
		}

		//QUERYSTRING: Deny high bit characters (shellcode)
		if(Settings.QueryString.HighBitShellcode){
			if(ScanHighBitShellCode(QueryString)){
				action.Set(Settings.QueryString.HighBitShellcode);
				Alert("High bit characters (shellcode) not allowed in querystring",Settings.QueryString.HighBitShellcode);
			}
		}

		//QUERYSTRING: Deny special whitespace characters
		if(Settings.QueryString.SpecialWhitespace){
			if(DenySpecialWhitespace(QueryString," not allowed in querystring",Settings.QueryString.SpecialWhitespace)){
				action.Set(Settings.QueryString.SpecialWhitespace);
			}
		}

		CStringHelper::Safe_MakeLower(QueryString);
		
		//QUERYSTRING: SQL injection (words in list should be lowercase)
		if(Settings.QueryString.SQLInjection){
			int sqlcount(0);
			CStringList Matches;
			sqlcount += ScanSQLInjection(QueryString,Settings.SQLi.Keywords,Matches);
			if(sqlcount>0){
				if(sqlcount>Settings.SQLi.AllowedCount){
					action.Set(Settings.QueryString.SQLInjection);
					Alert("Possible SQL injection in querystring: " + CStringHelper::ToString(Matches,","),Settings.QueryString.SQLInjection);
				}else{
					logger.Append("WARNING: SQL keyword found in querystring (" + CStringHelper::Safe_IntToAscii(Settings.SQLi.AllowedCount - sqlcount + 1) + " more will block request): " + CStringHelper::ToString(Matches,","));
				}
			}
		}

		//QUERYSTRING: deny certain sequences (list should be lowercase)
		if(Settings.QueryString.DenySequences.Action){
			if(DenySequences(QueryString,Settings.QueryString.DenySequences.List,"'","' not allowed in querystring",Settings.QueryString.DenySequences.Action)){
				action.Set(Settings.QueryString.DenySequences.Action);
			}
		}

		//QUERYSTRING: deny regex
		if(Settings.QueryString.DenyRegex.Action){
			if(DenyRegex(QueryString,Settings.QueryString.DenyRegex.Map,"'","' regex not allowed in data",Settings.QueryString.DenyRegex.Action)){
				action.Set(Settings.QueryString.DenyRegex.Action);
			}
		}

		return action;
	}
}

Action CHTTPFirewall::ScanConnection(CString &IP, bool& is_blacklisted, bool& is_monitored)
{
	Action action;

	//CONNECTION - IP blocking
	if(Settings.Connection.DenyAddresses.Enabled){
		if(Settings.IPRanges.Deny.IsInList(IP)){
			is_blacklisted = true;
			action.Set(Settings.Connection.DenyAddresses.Enabled);
			Alert("IP address '" + IP + "' not allowed",Settings.Connection.DenyAddresses.Enabled);
		}
	}

	//CONNECTION - Request Limits
	if(Settings.Connection.RequestsLimit.Action){
		//add request to cache
		Statistics.IP_RequestsLimit.Add(IP);

		//block if number of requests exceeds certain treshold
		if(Statistics.IP_RequestsLimit.Count(IP)>Settings.Connection.RequestsLimit.MaxCount){
			action.Set(Settings.Connection.RequestsLimit.Action);
			Alert("Reached connection request limit",Settings.Connection.RequestsLimit.Action);
		}
	}

	//CONNECTION - Hacking IP Blocking (do not check for Settings.HTTPResponse.BlockIP.Enabled - see rule action)
	if(Statistics.IP_HACK_Blocked.Count(IP)>Settings.HTTPResponse.BlockIP.MaxCount){
		is_blacklisted = true;
		action.Set(true);/* Settings.HTTPResponse.BlockIP.Enabled can be false*/
		Alert("Blacklisted IP address (previous alert)",true /* Settings.HTTPResponse.BlockIP.Enabled can be false*/);
	}

	//RESPONSE MONITOR
	if(Settings.ResponseMonitor.ClientErrors.Action){
		if(Statistics.IP_RESPONSE_Client_Errors.Count(IP)>Settings.ResponseMonitor.ClientErrors.MaxCount){
			is_blacklisted = true;
			action.Set(Settings.ResponseMonitor.ClientErrors.Action);
			Alert("Blacklisted IP address (multiple HTTP client errors)",Settings.ResponseMonitor.ClientErrors.Action);
		}
	}
	if(Settings.ResponseMonitor.ServerErrors.Action){
		if(Statistics.IP_RESPONSE_Server_Errors.Count(IP)>Settings.ResponseMonitor.ServerErrors.MaxCount){
			is_blacklisted = true;
			action.Set(Settings.ResponseMonitor.ServerErrors.Action);
			Alert("Blacklisted IP address (multiple HTTP server errors)",Settings.ResponseMonitor.ServerErrors.Action);
		}
	}

	//CONNECTION - Hacking IP Monitoring (do not check for Settings.HTTPResponse.MonitorIP - see rule action)
	if(Statistics.IP_HACK_Monitored.Exists(IP)){
		is_monitored = true;
		logger.Append("MONITORED: IP address (previous alert)");
	}

	//CONNECTION - IP Monitoring
	if(Settings.Connection.MonitorAddresses.Enabled){
		if(Settings.IPRanges.Monitor.IsInList(IP)){
			is_monitored = true;
			logger.Append("MONITORED: IP address '" + IP + "'");
		}
	}

	return action;
}

Action CHTTPFirewall::ScanAuthentication(CString &User, CString &Pass, CString& IP)
{
	Action action;

	//AUTH - block blank password
	if(Settings.Authentication.BlankPasswords && Pass==""){
		action.Set(Settings.Authentication.BlankPasswords);
		Alert("Blank passwords not allowed",Settings.Authentication.BlankPasswords);
	}
	//AUTH - block password=username
	if(Settings.Authentication.SamePasswordAsUsername && Pass.CompareNoCase(User)==0){
		action.Set(Settings.Authentication.SamePasswordAsUsername);
		Alert("Password is same as username",Settings.Authentication.SamePasswordAsUsername);
	}
	//AUTH - block default password lists
	if(Settings.Authentication.DenyDefaultPasswords.Action){
		if(Settings.Authentication.DenyDefaultPasswords.List.Find(Pass)!=NULL){
			action.Set(Settings.Authentication.DenyDefaultPasswords.Action);
			Alert("Default password '" + Pass + "' not allowed",Settings.Authentication.DenyDefaultPasswords.Action);
		}
	}
	//TODO: AUTH - require complex password (use regex)
	
	//AUTH - Scan Account
	action |= ScanAccount(User);

	CStringHelper::Safe_MakeUpper(User);
		
	//AUTH - scan for DoS on authentication of accounts (lookup)
	if(Settings.Authentication.BruteForceAttack.Action){
		CString ipuser; ipuser=IP+User;
		//add this ip to cache
		if(Statistics.IP_Authentication.Count(ipuser,Pass)==0){
			Statistics.IP_Authentication.Add(ipuser,Pass);
		}
		//count ip references in cache
		if(Statistics.IP_Authentication.Count(ipuser)>Settings.Authentication.BruteForceAttack.MaxCount){
			action.Set(Settings.Authentication.BruteForceAttack.Action);
			Alert("Possible brute force attack or DoS.",Settings.Authentication.BruteForceAttack.Action);
		}
	}

	return action;
}

Action CHTTPFirewall::ScanVerb(CString Verb)
{
	Action action;
	logger.Append(Verb.Left(2048));

	//VERB: only allow certain verbs (verbs in list are case sensitive!)
	if(Settings.Verbs.AllowVerbs.Action){
		if(!IsInList(Verb,Settings.Verbs.AllowVerbs.List)){
			action.Set(Settings.Verbs.AllowVerbs.Action);
			Alert("HTTP VERB not allowed",Settings.Verbs.AllowVerbs.Action);
		}
	}
	
	//VERB: Deny certain verbs (verbs in list should be uppercase)
	CStringHelper::Safe_MakeUpper(Verb); //Also deny lowercase & variants!
	if(Settings.Verbs.DenyVerbs.Action){
		if(IsInList(Verb,Settings.Verbs.DenyVerbs.List)){
			action.Set(Settings.Verbs.DenyVerbs.Action);
			Alert("HTTP VERB not allowed",Settings.Verbs.DenyVerbs.Action);
		}
	}

	return action;
}

Action CHTTPFirewall::ScanRawURL(CURL &RawUrl, CString& IP)
{
	Action action;
	CString URLDecoded = CURL::Decode(RawUrl.Url);

	//URL: RFC compliant
	if(Settings.URL.RFCCompliantURL){
		if(!RawUrl.IsRFCCompliant()){
			logger.Append(RawUrl.RawUrl.Left(4096));
			action.Set(Settings.URL.RFCCompliantURL);
			Alert("URL is not RFC compliant!",Settings.URL.RFCCompliantURL);
		}
	}

	//URL: HTTP RFC compliant
	if(Settings.URL.RFCCompliantHTTPURL){
		if(!RawUrl.IsHTTPCompliantUrl()){
			logger.Append(RawUrl.RawUrl.Left(4096));
			action.Set(Settings.URL.RFCCompliantHTTPURL);
			Alert("URL is not not RFC compliant (HTTP URL cannot contain authentication)!",Settings.URL.RFCCompliantHTTPURL);
		}
	}

	//URL
	logger.Append(RawUrl.Url.Left(4096));

	//URL Scanning (RAW)
	if(Settings.URL.UseRawScan){

		//URL: Check size
		if(Settings.URL.MaxLength.Action){
			if (RawUrl.Url.GetLength() > Settings.URL.MaxLength.Value){
				action.Set(Settings.URL.MaxLength.Action);
				Alert("URL is too long!",Settings.URL.MaxLength.Action);
			}
		}

		//URL: check for encoding abuse
		if(Settings.URL.EncodingExploits){
			if(ScanHexEncodingAbuse(RawUrl.Url,Settings.EncodingExploits.Keywords)){
				action.Set(Settings.URL.EncodingExploits);
				Alert("Encoding exploit in URL",Settings.URL.EncodingExploits);
			}

			//URL: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
			if(Settings.EncodingExploits.ScanDoubleEncoding && CanHexDecode(URLDecoded,false)){
				action.Set(Settings.URL.EncodingExploits);
				Alert("URL embedded encoding exploit",Settings.URL.EncodingExploits);
			}
		}

		//URL: scan
		action |= ScanURL(URLDecoded);
		action |= ScanURLDoS(URLDecoded);
	}

	//FILENAME (RAW)
	if(Settings.Filename.UseRawScan){
		CFileName fn(URLDecoded);
		action |= ScanFilename(fn);

		if(Settings.Global.IsInstalledAsGlobalFilter){ //no OnUrlMap in ISA/TMG
			action |= ScanExtensionDoS(fn.ExtensionLCase,IP);
		}
	}

	//QUERYSTRING (RAW)
	if(Settings.QueryString.UseRawScan && RawUrl.Querystring!=""){
		action |= ScanRawQuerystring(RawUrl.Querystring);
	}

	return action;
}

Action CHTTPFirewall::ScanVersion(const CString &Version, const CString& Host)
{
	Action action;
	logger.Append(Version.Left(1024));

	//VERSION: Check size
	if(Settings.Limits.MaxHTTPVersion.Action){
		if (Version.GetLength() > Settings.Limits.MaxHTTPVersion.Value){
			action.Set(Settings.Limits.MaxHTTPVersion.Action);
			Alert("HTTP version too long",Settings.Limits.MaxHTTPVersion.Action);
		}
	}
		
	//VERSION: allowed HTTP versions
	if(Settings.Limits.AllowHTTPVersions.Action){
		if(!IsInList(Version,Settings.Limits.AllowHTTPVersions.List)){
			action.Set(Settings.Limits.AllowHTTPVersions.Action);
			Alert("HTTP version not in allowed list",Settings.Limits.AllowHTTPVersions.Action);
		}
	}

	//denied http versions not needed
	
	//VERSION: host header required for HTTP 1.1 (most likely worms/script kiddies/other stuff that is not RFC compliant)
	if(Settings.Headers.RFCCompliantHostHeader){
		if(Version=="HTTP/1.1" || Version=="HTTP/2.0"){
			if(Host==""){
				action.Set(Settings.Headers.RFCCompliantHostHeader);
				Alert("'Host:' header required for HTTP 1.1 request",Settings.Headers.RFCCompliantHostHeader);
			}
			/*
		}else{
			if((Version=="HTTP/0.9" || Version=="") && Host!=""){
				action.Set(Settings.Headers.RFCCompliantHostHeader);
				Alert("'Host:' header not allowed in HTTP 0.9",Settings.Headers.RFCCompliantHostHeader);
			}//*/
		}
	}

	return action;
}

Action CHTTPFirewall::ScanCookie(CString &Cookie)
{
	Action action;
	if(Cookie.GetLength()==0){
		return action;
	}else{

		logger.Append(Cookie);

		//COOKIE: variable size
		if(Settings.Cookie.MaxVariableLength.Action && Settings.Cookie.MaxVariableLength.Value>0){
			if(ScanVariableLength(Cookie,Settings.Cookie.MaxVariableLength.Value,' ',' ')){
				action.Set(Settings.Cookie.MaxVariableLength.Action);
				Alert("Maximum variable length exceeded in cookie",Settings.Cookie.MaxVariableLength.Action);
			}
		}
	
		//COOKIE: check for encoding abuse
		if(Settings.Cookie.EncodingExploits){
			if(ScanHexEncodingAbuse(Cookie,Settings.EncodingExploits.Keywords)){
				action.Set(Settings.Cookie.EncodingExploits);
				Alert("Encoding exploit in cookie",Settings.Cookie.EncodingExploits);
			}
		}

		Cookie = CURL::Decode(Cookie); //can be encoded: http://wp.netscape.com/newsref/std/cookie_spec.html

		//COOKIE: Deny encoding exploits (normalize twice, except for '+', and reject if change occurs)
		if(Settings.Cookie.EncodingExploits && Settings.EncodingExploits.ScanDoubleEncoding){
			if(CanHexDecode(Cookie,false)){
				action.Set(Settings.Cookie.EncodingExploits);
				Alert("Encoding exploit in cookie (embedded encoding)",Settings.Cookie.EncodingExploits);
			}
		}

		//COOKIE: Deny directory traversal (parent path)
		if(Settings.Cookie.DirectoryTraversal){
			if(ScanDirectoryTraversal(Cookie)){
				action.Set(Settings.Cookie.DirectoryTraversal);
				Alert("Directory traversal not allowed in cookie",Settings.Cookie.DirectoryTraversal);
			}
		}

		//COOKIE: Deny high bit characters (shellcode)
		if(Settings.Cookie.HighBitShellcode){
			if(ScanHighBitShellCode(Cookie)){
				action.Set(Settings.Cookie.HighBitShellcode);
				Alert("High bit characters (shellcode) not allowed in cookie",Settings.Cookie.HighBitShellcode);
			}
		}

		//COOKIE: Deny special whitespace characters
		if(Settings.Cookie.SpecialWhitespace){
			if(DenySpecialWhitespace(Cookie," not allowed in cookie",Settings.Cookie.SpecialWhitespace)){
				action.Set(Settings.Cookie.SpecialWhitespace);
			}
		}

		CStringHelper::Safe_MakeLower(Cookie);

		//COOKIE: Deny SQL injection
		if(Settings.Cookie.SQLInjection){
			int sqlcount( Settings.SQLi.Keywords.Find(";")!=NULL?-1:0 );
			CStringList Matches;
			sqlcount += ScanSQLInjection(Cookie,Settings.SQLi.Keywords,Matches);
			if(sqlcount>0){
				if(sqlcount>Settings.SQLi.AllowedCount){
					action.Set(Settings.Cookie.SQLInjection);
					Alert("Possible SQL injection in cookie: " + CStringHelper::ToString(Matches,","),Settings.Cookie.SQLInjection);
				}else{
					logger.Append("WARNING: SQL keyword found in cookie (" + CStringHelper::Safe_IntToAscii(Settings.SQLi.AllowedCount - sqlcount + 1) + " more will block request): " + CStringHelper::ToString(Matches,","));
				}
			}			
		}

		//COOKIE: deny certain sequences (list should be lowercase)
		if(Settings.Cookie.DenySequences.Action){
			if(DenySequences(Cookie,Settings.Cookie.DenySequences.List,"'","' not allowed in cookie",Settings.Cookie.DenySequences.Action)){
				action.Set(Settings.Cookie.DenySequences.Action);
			}
		}

		return action;
	}
}

Action CHTTPFirewall::ScanPath(CFileName& fn)
{
	Action action;
	logger.Append(fn.FullPath.Left(9000));

	//FULLPATH
	if(Settings.Path.ParentPath){
		if(DenyString(fn.FullPath,"..","Parent Path attempt ('..')",Settings.Path.ParentPath)){
			action.Set(Settings.Path.ParentPath);
		}
	}

	if(Settings.Path.Escaping){
		if(DenyChar(fn.FullPath,'%',"Possible encoding exploit",Settings.Path.Escaping)){
			action.Set(Settings.Path.Escaping);
		}
	}

	if(Settings.Path.SpecialWhitespace){
		if(DenySpecialWhitespace(fn.FullPath," not allowed in path",Settings.Path.SpecialWhitespace)){
			action.Set(Settings.Path.SpecialWhitespace);
		}
	}

	if(Settings.Path.MultipleColons){
		if(DenyMultipleCharacters(fn.FullPath,':',"Multiple colons not allowed in path",Settings.Path.MultipleColons)){
			action.Set(Settings.Path.MultipleColons);
		}
	}

	if(Settings.Path.Characters.Action){
		char c;
		for(int i=0;i<Settings.Path.Characters.Value.GetLength();i++){
			c = Settings.Path.Characters.Value.GetAt(i);
			if(DenyChar(fn.FullPath,c,"'" + CString(c) + "' not allowed in path",Settings.Path.Characters.Action))
				action.Set(Settings.Path.Characters.Action);
		}
	}

	//PATH: Deny dot in path (!= full path)
	if(Settings.Path.Dot){
		if(DenyChar(fn.Path,'.',"'.' not allowed in path",Settings.Path.Dot)){
			action.Set(Settings.Path.Dot);
		}
	}
	
	//PATH: Only allow certain paths (I use full paths here, because when
	//      using path, virtual dirs don't work if they are set to c:\somedir
	//      instead of c:\somedir\											*/
	if(Settings.Path.AllowPaths.Action){ //(lowercase!)
		if(!IsStartInList(fn.FullPathLCase,Settings.Path.AllowPaths.List)){
			action.Set(Settings.Path.AllowPaths.Action);
			Alert("Not in allowed path list '" + fn.FullPath + "'",Settings.Path.AllowPaths.Action);
		}
	}

	//FILENAME
	action |= ScanFilename(fn);

	return action;
}

Action CHTTPFirewall::ScanExtensionDoS(CString& ExtensionLCase, CString& IP)
{
	Action action;
	//EXTENSION - Request Limits
	if(Settings.Extensions.RequestsLimit.Action && IsInList(ExtensionLCase,Settings.Extensions.LimitExtensions)){
		//add request to cache
		Statistics.Extension_RequestsLimit.Add(IP);

		//block if number of requests exceeds certain treshold
		if(Statistics.Extension_RequestsLimit.Count(IP)>Settings.Extensions.RequestsLimit.MaxCount){
			action.Set(Settings.Extensions.RequestsLimit.Action);
			Alert("Reached file extension requests limit",Settings.Extensions.RequestsLimit.Action);
		}
	}
	return action;
}

Action CHTTPFirewall::ScanUserAgent(CString &UserAgent, CString& IP)
{
	Action action;

	//User-Agent 		=	"User-Agent" ":" 1*( product | comment )
	//product 			=	token ["/" product-version ]
	//product-version 	=	token
	//comment 			=	"(" *( ctext | comment ) ")"
	//ctext 			=	<any TEXT excluding "(" and ")">
	//token 			=	1*<any CHAR except CTLs or tspecials>
	//tspecials 		=	"(" | ")" | "<" | ">" | "@" | "," | ";" | ":" | "\" | <"> | "/" | "[" | "]" | "?" | "=" | "{" | "}" | SP | HT

	//multiple User-Agents (or any headers exception):
	//Multiple message-header fields with the same field-name MAY be present in a message
	//if and only if the entire field-value for that header field is defined as a comma-separated
	//list [i.e., #(values)]. It MUST be possible to combine the multiple header fields into
	//one "field-name: field-value" pair, without changing the semantics of the message, by
	//appending each subsequent field-value to the first, each separated by a comma.

	if(UserAgent.GetLength()==0){
		//UA: Deny empty
		if(Settings.UserAgent.Empty){
			action.Set(Settings.UserAgent.Empty);
			Alert("Empty User Agent not allowed",Settings.UserAgent.Empty);
		}
	}else{

		//UA: require certain characters ('/ -.')
		if(Settings.UserAgent.RequireCharacter.Action && Settings.UserAgent.RequireCharacter.Value.GetLength()>0){
			if(UserAgent.FindOneOf(Settings.UserAgent.RequireCharacter.Value)==-1){
				action.Set(Settings.UserAgent.RequireCharacter.Action);
				Alert("User Agent requires at least one of these characters: '" + Settings.UserAgent.RequireCharacter.Value + "'",Settings.UserAgent.RequireCharacter.Action);
			}
		}

		//UA: Deny User Agent not RFC compliant
		if(Settings.UserAgent.NonRFC){
			//remove comments
			CString UserAgentClean = RemoveRFCComments(UserAgent);
			//logger.Append(UserAgentClean);
			//scan for non-token chars
			//if(UserAgentClean.FindOneOf("()<>@,;:\\\"/[]?={} \t")!=-1){
			if(UserAgentClean.FindOneOf("()<>@;:\\\"[]?={}\t")!=-1){	//allow '/' + ' ' + ',' (multiple headers with same name)
				action.Set(Settings.UserAgent.NonRFC);
				Alert("User Agent not RFC compliant",Settings.UserAgent.NonRFC);
			}
		}

		//UA: Deny high bit characters (shellcode)
		if(Settings.UserAgent.HighBitShellcode){
			if(ScanHighBitShellCode(UserAgent)){
				action.Set(Settings.UserAgent.HighBitShellcode);
				Alert("High bit characters (shellcode) not allowed in User Agent",Settings.UserAgent.HighBitShellcode);
			}
		}

		//UA: Deny special whitespace characters
		if(Settings.UserAgent.SpecialWhitespace){
			if(DenySpecialWhitespace(UserAgent," not allowed in User Agent",Settings.UserAgent.SpecialWhitespace)){
				action.Set(Settings.UserAgent.SpecialWhitespace);
			}
		}

		//UA: Deny switching User-Agent
		if(Settings.UserAgent.Switching.Action){
			//add to cache
			Statistics.IP_UserAgent.AddNotInCache(IP,UserAgent);

			//block if number of requests exceeds certain treshold
			if(Statistics.IP_UserAgent.Count(IP)>Settings.UserAgent.Switching.MaxCount){
				action.Set(Settings.UserAgent.Switching.Action);
				Alert("User Agent is switching too much on this IP address",Settings.UserAgent.Switching.Action);
			}
		}

		//UA: Denied User-Agents (case insensitive)
		if(Settings.UserAgent.DenyUserAgents.Action){
			if(IsInListCompareNoCase(UserAgent,Settings.UserAgent.DenyUserAgents.List)){
				action.Set(Settings.UserAgent.DenyUserAgents.Action);
				Alert("User Agent not allowed",Settings.UserAgent.DenyUserAgents.Action);
			}
		}

		//UA: Denied User-Agent Sequences (list should be lowercase)
		if(Settings.UserAgent.DenySequences.Action){
			CStringHelper::Safe_MakeLower(UserAgent);
			if(DenySequences(UserAgent,Settings.UserAgent.DenySequences.List,"'","' not allowed in User Agent",Settings.UserAgent.DenySequences.Action)){
				action.Set(Settings.UserAgent.DenySequences.Action);
			}
		}
	}

	return action;
}

Action CHTTPFirewall::ScanContent(CString &ContentType, CString &ContentLength)
{
	Action action;
	int l(0);

	//HEADERS: Check Content-Length value
	if((l = ContentLength.GetLength())!=0 && Settings.Limits.MaxContentLength.Action){
		//to prevent overflow or other types of attack, block request if the length
		//of Content-Length header is longer than length of max allowed value
		//or longer than 10 characters (maximum value of long integer)!
		if(l > Settings.Limits.MaxContentLength.Value.GetLength() || l > 10 || 
			(l == 10 && !(ContentLength.Left(1)=="1" || 
							(ContentLength.Left(1) == "2" && (	ContentLength.Left(2)=="20" || 
																ContentLength.Left(3)=="210" || 
																ContentLength.Left(3)=="211" || 
																ContentLength.Left(3)=="212" || 
																ContentLength.Left(3)=="213"
															 )	
							)
						 )
			)
		){
			action.Set(Settings.Limits.MaxContentLength.Action);
			Alert("Content-Length is too long (size)!",Settings.Limits.MaxContentLength.Action);
			logger.Append(ContentLength.Left(1024));
		}else{
			if (atol(ContentLength) > Settings.Limits.lMaxContentLength){
				action.Set(Settings.Limits.MaxContentLength.Action);
				Alert("Content-Length is too long (value)!",Settings.Limits.MaxContentLength.Action);
				logger.Append(ContentLength);
			}
		}
	}

	//HEADERS: Allowed Content-Types
	if(Settings.ContentType.AllowContentTypes.Action){
		if(!IsNotEmptyStartInList(ContentType,Settings.ContentType.AllowContentTypes.List)){
			action.Set(Settings.ContentType.AllowContentTypes.Action);
			Alert("Content-Type '" + ContentType.Left(2048) + "' not allowed",Settings.ContentType.AllowContentTypes.Action);
		}
	}

	//HEADERS: Denied Content-Types
	if(Settings.ContentType.DenyContentTypes.Action && ContentType.GetLength()>0){
		if(IsStartInList(ContentType,Settings.ContentType.DenyContentTypes.List)){
			action.Set(Settings.ContentType.DenyContentTypes.Action);
			Alert("Content-Type '" + ContentType.Left(2048) + "' not allowed",Settings.ContentType.DenyContentTypes.Action);
		}
	}
	
	return action;
}

bool CHTTPFirewall::IsExcludedSite(CString &instance)
{
	if(Settings.Engine.ExcludedInstances.Enabled && IsInList(instance,Settings.Engine.ExcludedInstances.List)){
		logger.Append("Web site instance is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedHost(CString& Host)
{
	if(Settings.Engine.ExcludedHosts.Enabled && IsInList(Host,Settings.Engine.ExcludedHosts.List)){
		logger.Append("Web site host is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedIP(CString& IP)
{
	if(Settings.Engine.ExcludedIPAddresses.Enabled && Settings.IPRanges.Exclude.IsInList(IP)){
		logger.Append("IP address is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedURL(CString& url, CString& Querystring)
{
	if	(
			(Settings.URL.ExcludedURL.Enabled && IsStartInList(url,Settings.URL.ExcludedURL.List)) ||
			(Settings.QueryString.ExcludedQueryString.Enabled && IsInList(Querystring,Settings.QueryString.ExcludedQueryString.List))
		)
	{
		logger.Append(CString(url + '?' + Querystring).Left(2048));
		logger.Append("URL is excluded from scanning");
		return true;
	}
	return false;
}

bool CHTTPFirewall::IsExcludedUserAgent(CString& UserAgent)
{
	if(Settings.UserAgent.Excluded.Enabled && IsInList(UserAgent,Settings.UserAgent.Excluded.List)){
		logger.Append("User Agent is excluded from scanning");
		return true;
	}
	return false;
}

void CHTTPFirewall::LoadCache()
{
	CFirewall::LoadCache();

	//Content Length
	Settings.Limits.lMaxContentLength = atol(Settings.Limits.MaxContentLength.Value);

	//Status code
	Settings.HTTPResponse.SyncStatusCode();

	//Response Server header
	Settings.ResponseMonitor.SyncServerHeader(true); //for backwards compatibility (when using old settings file)
	Settings.ResponseMonitor.SyncServerHeader(); //update ServerHeader variable...

	//(re-)load IP Lists
	Settings.IPRanges.Monitor.Enumerate(Settings.Connection.MonitorAddresses.List);
	Settings.IPRanges.Deny.Enumerate(Settings.Connection.DenyAddresses.List);
	Settings.IPRanges.Exclude.Enumerate(Settings.Engine.ExcludedIPAddresses.List);
	Settings.IPRanges.Admin.Enumerate(Settings.Admin.FromIP);

	//stringcaches
	Statistics.IP_Authentication.SetMaxAge(CTimeSpan(0,0,Settings.Authentication.BruteForceAttack.MaxTime>1?Settings.Authentication.BruteForceAttack.MaxTime:1,0));//in min.
	Statistics.IP_RequestsLimit.SetMaxAge(CTimeSpan(0,0,Settings.Connection.RequestsLimit.MaxTime>1?Settings.Connection.RequestsLimit.MaxTime:1,0));	//in min.

	Statistics.IP_HACK_Monitored.SetMaxAge(CTimeSpan(0,Settings.HTTPResponse.MonitorIPTimeout>1?Settings.HTTPResponse.MonitorIPTimeout:1,0,0));//in hours.
	Statistics.IP_HACK_Blocked.SetMaxAge(CTimeSpan(0,Settings.HTTPResponse.BlockIP.MaxTime>1?Settings.HTTPResponse.BlockIP.MaxTime:1,0,0));//in hours.

	Statistics.URL_RequestsLimit.SetMaxAge(CTimeSpan(0,0,0,Settings.URL.RequestsLimit.MaxTime>1?Settings.URL.RequestsLimit.MaxTime:1));	//in sec.
	Statistics.Extension_RequestsLimit.SetMaxAge(CTimeSpan(0,0,Settings.Extensions.RequestsLimit.MaxTime>1?Settings.Extensions.RequestsLimit.MaxTime:1,0));	//in min.
	Statistics.IP_UserAgent.SetMaxAge(CTimeSpan(0,0,Settings.UserAgent.Switching.MaxTime>1?Settings.UserAgent.Switching.MaxTime:1,0));	//in min.

	Statistics.IP_RESPONSE_Client_Errors.SetMaxAge(CTimeSpan(0,0,Settings.ResponseMonitor.ClientErrors.MaxTime>1?Settings.ResponseMonitor.ClientErrors.MaxTime:1,0));	//in min.
	Statistics.IP_RESPONSE_Server_Errors.SetMaxAge(CTimeSpan(0,0,Settings.ResponseMonitor.ServerErrors.MaxTime>1?Settings.ResponseMonitor.ServerErrors.MaxTime:1,0));	//in min.
}

void CHTTPFirewall::ClearCache()
{
	CFirewall::ClearCache();
	Statistics.Clear();
}

void CHTTPFirewall::SetLoggerFields()
{
	//set logger fields
	CStringList fields;
	fields.AddTail("Date");
	fields.AddTail("Time");
	fields.AddTail("Site Instance");
	fields.AddTail("Event");
	if(Settings.Logging.LogClientIP) fields.AddTail("Client IP");
	if(Settings.Logging.LogUserName) fields.AddTail("Username");
	if(Settings.HTTPLogging.Log_HTTP_VIA) fields.AddTail("Via header");
	if(Settings.HTTPLogging.Log_HTTP_X_FORWARDED_FOR) fields.AddTail("HTTP_X_FORWARDED_FOR");
	if(Settings.HTTPLogging.LogHostHeader) fields.AddTail("Host header");
	if(Settings.HTTPLogging.LogUserAgent) fields.AddTail("User agent");
	fields.AddTail("Additional info about request (event specific)");
	logger.SetHeaderFields(fields);
}

Action CHTTPFirewall::ScanRawRefererURL(LPCTSTR ReferrerUrl,LPCTSTR RequestedURLDecoded, CString& HostHeader)
{
	CURL RefUrl(ReferrerUrl);
	CFileName fn(RequestedURLDecoded);
	return ScanRawRefererURL(RefUrl,fn,HostHeader);
}

Action CHTTPFirewall::ScanRawRefererURL(CURL &ReferrerUrl, CFileName& RequestedFile, CString& HostHeader)
{
	Action action;

	if(Settings.Referrer.UseScan){

		CString URLDecoded = CURL::Decode(ReferrerUrl.Url);

		//REFERER URL
		logger.Append(ReferrerUrl.RawUrl.Left(4096));
		
		//REFERER URL: RFC compliant
		if(Settings.Referrer.RFCCompliantURL){
			if(!ReferrerUrl.IsRFCCompliant()){
				action.Set(Settings.Referrer.RFCCompliantURL);
				Alert("Referrer URL is not RFC compliant!",Settings.Referrer.RFCCompliantURL);
			}
		}

		//REFERER URL: HTTP RFC compliant
		if(Settings.Referrer.RFCCompliantHTTPURL){
			if(!ReferrerUrl.IsHTTPCompliantUrl()){
				action.Set(Settings.Referrer.RFCCompliantHTTPURL);
				Alert("Referrer URL is not not RFC compliant (HTTP URL cannot contain authentication)!",Settings.Referrer.RFCCompliantHTTPURL);
			}
		}

		//REFERER URL: check for encoding abuse & exploits
		if(Settings.Referrer.EncodingExploits){
			//REFERER URL: check for encoding abuse
			if(ScanHexEncodingAbuse(ReferrerUrl.Url,Settings.EncodingExploits.Keywords)){
				action.Set(Settings.Referrer.EncodingExploits);
				Alert("Encoding exploit in Referer URL",Settings.Referrer.EncodingExploits);
			}

			//REFERER URL: Encoding Exploits (normalize twice, except for '+', and reject if change occurs)
			if(Settings.EncodingExploits.ScanDoubleEncoding && CanHexDecode(URLDecoded,false)){
				action.Set(Settings.Referrer.EncodingExploits);
				Alert("Referrer URL embedded encoding exploit",Settings.Referrer.EncodingExploits);
			}
		}

		//REFERER URL: scan
		action |= ScanRefererURL(URLDecoded,RequestedFile,HostHeader);

		/*
		//FILENAME (RAW)
		if(Settings.UseFilenameRawScan){
			action |= ScanFilename(CFileName(URLDecoded));
		}

		//QUERYSTRING (RAW)
		if(Settings.UseQuerystringRawScan && ReferrerUrl.Querystring!=""){
			action |= ScanRawQuerystring(ReferrerUrl.Querystring);
		}//*/
	}

	return action;
}

Action CHTTPFirewall::ScanRefererURL(CString& ReferrerURL, CFileName& RequestedFile, CString& HostHeader)
{
	Action action;

	//REFERER URL: Deny deep linking
	if(Settings.Referrer.HotLinking){
		//is certain url or file extension?
		if(IsListItemInString(RequestedFile.FullPathLCase,Settings.Referrer.HotLinkingUrls) || IsInList(RequestedFile.ExtensionLCase,Settings.Referrer.HotLinkingFileExtensions)){
			CString RefDomain = CURL::FQDNNoEndDot(ReferrerURL,false);
			CStringHelper::Safe_MakeLower(RefDomain);
			if(ReferrerURL.GetLength()>0){
				if(Settings.Referrer.HotLinkingUseHostHeader && RefDomain.CompareNoCase(HostHeader)==0 ){
					//allow: HostHeader == RefDomain
				}else{
					//REFERRER: No Host header
					if(Settings.Referrer.HotLinkingUseHostHeader && HostHeader.GetLength()==0){
						logger.Append("WARNING: Hot linking 'Host:' header missing!");
					}
					//REFERRER: Allowed domains (list should be lowercase)
					if(Settings.Referrer.HotLinkingAllowDomains.Action){
						if(!IsInList(RefDomain,Settings.Referrer.HotLinkingAllowDomains.List)){
							action.Set(Settings.Referrer.HotLinkingAllowDomains.Action);
							Alert("Hot linking domain not in allowed list",Settings.Referrer.HotLinkingAllowDomains.Action);
						}
					}
					//REFERRER: Denied domains (list should be lowercase)
					if(Settings.Referrer.HotLinkingDenyDomains.Action){
						if(IsInList(RefDomain,Settings.Referrer.HotLinkingDenyDomains.List)){
							action.Set(Settings.Referrer.HotLinkingDenyDomains.Action);
							Alert("Hot linking domain in denied list",Settings.Referrer.HotLinkingDenyDomains.Action);
						}
					}
				}
			}else{
				//REFERRER: Blank header
				if(Settings.Referrer.HotLinkingBlankReferrer){
					//also blocks some proxy servers and browsers with certain security applications
					action.Set(Settings.Referrer.HotLinkingBlankReferrer);
					Alert("Hot linking with blank referrer not allowed",Settings.Referrer.HotLinkingBlankReferrer);
				}
			}
		}
	}

	//REFERER URL: Deny certain characters
	if(Settings.Referrer.Characters.Action){
		char c;
		for(int i=0;i<Settings.Referrer.Characters.Value.GetLength();i++){
			c = Settings.Referrer.Characters.Value.GetAt(i);
			if(DenyChar(ReferrerURL,c,"'" + CString(c) + "' not allowed in Referrer URL",Settings.Referrer.Characters.Action))
				action.Set(Settings.Referrer.Characters.Action);
		}
	}

	//REFERER URL: Deny high bit characters (shellcode)
	if(Settings.Referrer.HighBitShellcode){
		if(ScanHighBitShellCode(ReferrerURL)){
			action.Set(Settings.Referrer.HighBitShellcode);
			Alert("High bit characters (shellcode) not allowed in Referrer URL",Settings.Referrer.HighBitShellcode);
		}
	}

	//REFERER URL: Deny special whitespace characters
	if(Settings.Referrer.SpecialWhitespace){
		if(DenySpecialWhitespace(ReferrerURL, " not allowed in Referrer URL",Settings.Referrer.SpecialWhitespace)){
			action.Set(Settings.Referrer.SpecialWhitespace);
		}
	}

	CStringHelper::Safe_MakeLower(ReferrerURL);

	//REFERER URL: Deny SQL injection
	if(Settings.Referrer.SQLInjection){
		int sqlcount(0);
		CStringList Matches;
		sqlcount += ScanSQLInjection(ReferrerURL,Settings.SQLi.Keywords,Matches);
		if(sqlcount>0){
			if(sqlcount>Settings.SQLi.AllowedCount){
				action.Set(Settings.Referrer.SQLInjection);
				Alert("Possible SQL injection in referrer URL: " + CStringHelper::ToString(Matches,","),Settings.Referrer.SQLInjection);
			}else{
				logger.Append("WARNING: SQL keyword found in referrer URL (" + CStringHelper::Safe_IntToAscii(Settings.SQLi.AllowedCount - sqlcount + 1) + " more will block request): " + CStringHelper::ToString(Matches,","));
			}
		}
	}

	//REFERER URL: Deny certain sequences (list should be lowercase)
	if(Settings.Referrer.DenySequences.Action){
		if(DenySequences(ReferrerURL,Settings.Referrer.DenySequences.List,"'","' not allowed in Referrer URL",Settings.Referrer.DenySequences.Action)){
			action.Set(Settings.Referrer.DenySequences.Action);
		}
	}

	return action;
}

CString CHTTPFirewall::GetDirectResponse(bool SendMessageBody, bool SendHeaders)
{
	if(Settings.HTTPResponse.CounterStrike){
		return GetInvalidResponse();
	}else{
		return GetAlertResponse(SendMessageBody, SendHeaders);
	}
}

CString CHTTPFirewall::GetAlertResponse(bool SendMessageBody, bool SendHeaders)
{
	CString buf("");
	if(SendHeaders){
		buf += "HTTP/1.1 ";
		if(Settings.HTTPResponse.UseStatus){
			buf += Settings.HTTPResponse.Status;
		}else{
			buf += "404 Object Not Found";
		}
		if(!Settings.ResponseMonitor.RemoveServerHeader){
			buf += "\r\nServer: ";
			if(Settings.ResponseMonitor.ChangeServerHeader){
				buf += Settings.ResponseMonitor.ServerHeader;
			}else{
				buf += Settings.Admin.ServerHeader;
			}
		}
		_tzset();
		CString d = CTime::GetCurrentTime().FormatGmt("%a, %d %b %Y %H:%M:%S GMT");
		buf += "\r\nDate: " + d;
		buf += "\r\nContent-Type: text/html; charset=windows-1252";
		buf += "\r\nContent-Length: "; buf += CStringHelper::Safe_IntToAscii(Settings.HTTPResponse.Contents.GetLength());
		buf += "\r\nPragma: no-cache";
		buf += "\r\nCache-control: no-cache";
		buf += "\r\nExpires: " + d;
		buf += CRLFCRLF;
	}

	if(SendMessageBody){ //see e-mail: special feature request of Sanford Whiteman if HEAD request -> no message body
		buf += Settings.HTTPResponse.Contents;
	}

	return buf;
}

CString CHTTPFirewall::GetInvalidResponse()
{
	/*
	 * This response tries to crash/close the client in several ways
	 */
	CString status;
	CString body;
	CString contentlength;
	switch(Settings.HTTPResponse.CounterStrikeId)
	{
		case 1: //Javascript close browser
		case 2: //twice for FF window.open below
			Settings.HTTPResponse.CounterStrikeId++;
			status = "HTTP/1.1 200 OK";
			body = "<html><head></head><body><a href=\"javascript:window.close();\">shutdown</a>";
			body += "<script>";
			//IE
			body += "window.open('', '_self', ''); window.close();";
			//FF - not working in versions > 21
			body += "window.open(document.URL,'_self','resizable=no,top=-245,width=250,height=250,scrollbars=no'); window.close();";
			//FF - go back in history
			body += "window.history.back(-5);";
			body += "</script></body></html>";
			contentlength = CStringHelper::Safe_IntToAscii(body.GetLength());
			break;
		case 3: //Javascript annoying message
			Settings.HTTPResponse.CounterStrikeId++;
			status = "HTTP/1.1 200 OK";
			body = "<html><head></head><body><a href=\"javascript:window.close();\">shutdown</a>";
			body += "<script>alert('No hacking allowed');</script></body></html>";
			contentlength = CStringHelper::Safe_IntToAscii(body.GetLength());
			break;
		case 4: //HTTP protocol issues
			Settings.HTTPResponse.CounterStrikeId++;
			status = "HTTP/1.9 99999999999999999999 No Hacking";
			contentlength = "99999999999999999999"; //larger than uint64 but same length to bypass stupid checks (18,446,744,073,709,551,615)
			body = Settings.HTTPResponse.Contents;
			break;
		case 5: //empty response
			Settings.HTTPResponse.CounterStrikeId++;
			return "";
		default: //normal message
			Settings.HTTPResponse.CounterStrikeId = 1;
			return GetAlertResponse(true);//always send body, even for HEAD requests (crashes certain tools)
	}
	CString buf(status);
	_tzset();
	CString d = CTime::GetCurrentTime().FormatGmt("%a, %d %b %Y %H:%M:%S GMT");
	buf += "\r\nDate: " + d;
	buf += "\r\nContent-Type: text/html; charset=windows-1252";
	buf += "\r\nContent-Length: "; buf += contentlength;
	buf += "\r\nPragma: no-cache";
	buf += "\r\nCache-control: no-cache";
	buf += "\r\nExpires: " + d;
	buf += CRLFCRLF;
	buf += body; //always send data, even for HEAD requests (crashes certain tools)
	return buf;
}

void CHTTPFirewall::HackResponseBlockOrMonitor(CString& IP)
{
	//block or monitor IP address
	if(Settings.HTTPResponse.MonitorIP){
		Statistics.IP_HACK_Monitored.AddNotInCache(IP);
	}
	if(Settings.HTTPResponse.BlockIP.Enabled){
		Statistics.IP_HACK_Blocked.Add(IP); //don't use AddNotInCache (need count of alerts)
	}
}

void CHTTPFirewall::Initialize(CString ProductName, LPCTSTR ProductFullName, LPCTSTR UserAgentsFile, CString ConfigurationPath, LPCTSTR LogSuffix)
{
	CString strinit; CString strinit2; CString strinit3;
	CString owner = CProcessHelper::GetCurrentUserName();

	if(owner!="" && CFileName::Exists(ConfigurationPath + "Upgrade." + owner + ".xml")){
		//OWNER - upgrade settings
		strinit = LoadSettingFile(ConfigurationPath,"Upgrade." + owner + ".xml");
		if(CFileName::DeleteFile(ConfigurationPath + "Upgrade." + owner + ".xml")){
			CFileName::DeleteFile(ConfigurationPath + ProductName + "." + owner + ".xml");
		}
		Settings.Upgrade();
		//save settings file after upgrade
		strinit2 = SaveDefaultSettingFile(ConfigurationPath,ProductName + "." + owner + ".xml");
	}else{
		if(owner!="" && CFileName::Exists(ConfigurationPath + ProductName + "." + owner + ".xml")){
			//OWNER - settings per application pool identity (IIS7 by default possible, IIS 6 possible if you use separate application pool + identity)
			strinit = LoadSettingFile(ConfigurationPath,ProductName + "." + owner + ".xml");
		}else{
			if(CFileName::Exists(ConfigurationPath + "Upgrade.xml")){
				//upgrade settings
				strinit = LoadSettingFile(ConfigurationPath,"Upgrade.xml");
				if(CFileName::DeleteFile(ConfigurationPath + "Upgrade.xml")){
					CFileName::DeleteFile(ConfigurationPath + ProductName + ".xml");
				}
				Settings.Upgrade();
			}else{
				//default settings files
				strinit = LoadSettingFile(ConfigurationPath,ProductName + ".xml",ProductName + ".ini","urlscan.ini");
			}
			//create a default settings file if it doesn't exist
			strinit2 = SaveDefaultSettingFile(ConfigurationPath,ProductName + ".xml");
		}
	}

	CString logdir;
#ifdef SYSTEM_LOG_BUILD
	CString windir = GetWindowsPath();
	logdir = windir + "\\System32\\LogFiles\\" + ProductName + "\\"; //for global filter
	#pragma message("***** This build will log under the LogFiles folder of your Windows NT installation! *****")
#else
	logdir = ConfigurationPath  + "LogFiles\\";				   //<dir of dll>\LogFiles
#endif
	strinit3 = InitializeLogger(logdir,ProductFullName,LogSuffix);

	Message("AQTRONIX " + ProductName + " loaded");
	Message(strinit);

	if(strinit2!="")
		Message(strinit2);
	if(strinit3!="")
		Message(strinit3);

	if(logger.CurrentLog.GetLastError()!=""){
		CString strError = logger.CurrentLog.GetLastError();
		Message("ERROR: " + strError);
		logger.ErrorEntry(strError,"If you are running multiple instances of " + ProductName + " then only the first loaded instance has access to the log file. Make sure each instance of " + ProductName + " has a unique log file, you can also enable Per Process Logging or Per Process Owner Logging.");
	}

	//are we in logging only mode? if so, warning!
	WarnLogonlyMode();

	Settings.ReadAgentsFile(ConfigurationPath + UserAgentsFile);

	LoadCache();

	CFileName fnLoaded = Settings.GetLoadedSettingsFileName();
	SaveRunningValues(fnLoaded.Path,fnLoaded.FileName);
}

CString CHTTPFirewall::GetAdmin(CString Message)
{
	CString ret = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\r\n<html>\r\n<head>\r\n<title>" + Settings.Admin.ServerHeader + "</title>\r\n<META name=\"ROBOTS\" content=\"NOINDEX, NOFOLLOW\">\r\n";
	
	//toggle display
	ret += " <script language=\"javascript\" type=\"text/javascript\">\r\n";
    ret += "    function ToggleDisplay(id) { ";
    ret += "        var divstyle = new String(); ";
    ret += "        divstyle = document.getElementById(id).style.display; ";
    ret += "        if (divstyle == \"block\" || divstyle == \"\") { ";
    ret += "            document.getElementById(id).style.display = \"none\"; ";
    ret += "        } ";
    ret += "        else { ";
    ret += "            document.getElementById(id).style.display = \"block\"; ";
    ret += "        } ";
    ret += "    } ";
	ret += " </script>\r\n";

	//TODO: ADMIN - jquery
	//ret += "<link rel=\"stylesheet\" href=\"./js/jquery-ui.css\" />";
	//ret += "<script src=\"./js/jquery-1.9.1.js\"></script>";
	//ret += "<script src=\"./js/jquery-ui.js\"></script>";
	//ret += "<script>$(function() {$( \"#accordion\" ).accordion();});</script>";

	////TODO: ADMIN - tabletools (uses jquery)
	////http://datatables.net/extras/tabletools/
	//ret += "<script type=\"text/javascript\" src=\"/DataTables-1.9.4/media/js/jquery.js\"></script>\r\n";
	//ret += "<script type=\"text/javascript\" src=\"/DataTables-1.9.4/media/js/jquery.dataTables.min.js\"></script>\r\n";
	ret += "</head>\r\n";

	//Body
	ret += "<body text=\"#000000\" vlink=\"#FF3300\" link=\"#FF3300\" bgcolor=\"#ffffff\"><font face=\"Verdana,Arial,Helvetica\" size=\"2\">\r\n";
	ret += "<h2>" + Settings.Admin.ServerHeader + "</h2>\r\n";

	//Admin
	ret += "<p><a href=\"?\">Home</a> - <a href=\"?Log\">Log</a> - <a href=\"?Settings\">Settings</a> - <a href=\"?Robots\">Robots</a> - <a href=\"?Diagnostics\">Diagnostics</a> - <a href=\"?Help\">Help</a></p>\r\n";
	ret += "<p><strong>" + Message + "</strong></p>\r\n";


	//Stats
	ret += "<h3>Statistics</h3>\r\n";
	ret += "<noscript><p><strong><font color=\"red\">This functionality requires JavaScript to be enabled!</font></strong></p></noscript>\r\n";
	//TODO: ADMIN - jquery
	//ret += "<div id=\"accordion\">";

	//Stringcaches
	ret += GetStatistics("Alerts",Settings.Statistics.Alerts);
	ret += GetStatistics("Messages",Settings.Statistics.Messages);
	ret += GetStatistics("IP_Authentication",Statistics.IP_Authentication, false);
	ret += GetStatistics("IP_RequestsLimit",Statistics.IP_RequestsLimit);
	ret += GetStatistics("IP_Hack_Monitored",Statistics.IP_HACK_Monitored);
	ret += GetStatistics("IP_Hack_Blocked",Statistics.IP_HACK_Blocked);
	ret += GetStatistics("IP_Client_Errors",Statistics.IP_RESPONSE_Client_Errors);
	ret += GetStatistics("IP_Server_Errors",Statistics.IP_RESPONSE_Server_Errors);
	ret += GetStatistics("IP_UserAgent",Statistics.IP_UserAgent);
	ret += GetStatistics("URL_RequestsLimit",Statistics.URL_RequestsLimit);
	ret += GetStatistics("Extension_RequestsLimit",Statistics.Extension_RequestsLimit);
	ret += GetStatisticsAdditionalCaches();
	//ret += "</div>";

	//Parsed IP ranges
	ret += "<h3>IP Ranges</h3>\r\n";
	if(Settings.Connection.MonitorAddresses.Enabled)
		ret += GetStatistics("Monitored",Settings.IPRanges.Monitor);
	if(Settings.Connection.DenyAddresses.Enabled)
		ret += GetStatistics("Denied",Settings.IPRanges.Deny);
	if(Settings.Engine.ExcludedIPAddresses.Enabled)
		ret += GetStatistics("Excluded",Settings.IPRanges.Exclude);

	//Regex patterns
	ret += "<h3>Regex Patterns</h3>\r\n";
	ret += GetStatistics("Compiled",Settings.regex_cache);

	//show server time (local/GMT)
	ret += "<h3>Firewall</h3>\r\n";
	ret += "<table>\r\n";
	ret += "<tbody>\r\n";
	ret += AddRow("Local server time",CTime::GetCurrentTime().Format("%Y-%m-%d %H:%M:%S"));
	ret += AddRow("GMT/UTC",CTime::GetCurrentTime().FormatGmt("%Y-%m-%d %H:%M:%S"));
	ret += AddRow("Running as account", CHTML::HTMLEncode(CProcessHelper::GetCurrentUserName()));
	ret += AddRow("In the process", CHTML::HTMLEncode(CProcessHelper::GetProcessPath()));
	CString sf = Settings.GetReadSettingsFileName();
	if(sf==""){
		ret += AddRow("Settings","<font color=\"red\">Default settings loaded</font>. Make sure the account above has read access on the settings files.");
	}else{
		ret += AddRow("Settings loaded from file", CHTML::HTMLEncode(Settings.GetReadSettingsFileName()));
		ret += "<tr><td>Robots database:</td><td>" + CHTML::HTMLEncode(Settings.Agents.AgentList.GetReadSettingsFileName()) + ", Last updated:";
		CTime t = Settings.Agents.AgentList.GetReadSettingsTimestamp();
		if(t < CTime::GetCurrentTime() - CTimeSpan(30,0,0,0)){
			ret += " <font color=\"red\">" + t.Format("%Y-%m-%d") + "</font>";
			if(Settings.Admin.AllowConfigChanges)
				ret += " -&gt; <a href=\"default.asp?update=Robots\">Update</a>";
		}else{
			ret += " <font color=\"green\">" + t.Format("%Y-%m-%d") + "</font>";
		}
		ret += "</td></tr>\r\n";
	}
	ret += "<tr><td>Logging:</td><td>";
	if(Settings.Logging.Enabled){
		if(logger.CurrentLog.GetLastError()!=""){
			ret += " <font color=\"red\">";
			ret += CHTML::HTMLEncode(logger.CurrentLog.GetLastError());
			ret += "</font>";
		}else{
			ret += CHTML::HTMLEncode(logger.CurrentLog.GetFileName());
		}
	}else{
		ret += "disabled";
	}
	ret += "</td></tr>\r\n";
	ret += "</tbody>\r\n";
	ret += "</table>\r\n";

	//admin access
	ret += "<p>This page is only accessible from these IP ranges:</p>\r\n";
	ret += GetStatistics("Admin",Settings.IPRanges.Admin);
	
	//end
	ret += "</font>\r\n";

	ret += "</body>\r\n</html>";

	return ret;	
}

CString CHTTPFirewall::GetStatistics(LPCTSTR Title, CStringCache& cache, bool ShowValue)
{
	return CollapsibleTable(Title,cache.ToString("|||", ShowValue),cache.Count());
}

CString CHTTPFirewall::GetStatistics(LPCTSTR Title, CIPRangeList& list)
{
	return CollapsibleTable(Title,list.ToString("|||"),list.Count());
}

CString CHTTPFirewall::GetStatistics(LPCTSTR Title, CMapStringToPtr& map)
{
	int count = map.GetCount();
	CString ret = "<p>";
	ret += Title;
	ret += " (" + CStringHelper::Safe_IntToAscii(count) + ")";
	if(count>0){
		ret += " <a href=\"javascript:ToggleDisplay(\'";
		ret += Title;
		ret += "\');\">+</a>";
	}
	ret += "</p>\r\n";

	CString k;
	void* p;
	POSITION pos = map.GetStartPosition();
	if(pos!=NULL){
		ret += "<div id=\"";
		ret += Title;
		ret += "\" style=\"display:none;\">\r\n";
		ret += "<table border=\"1\">\r\n";
		while(pos!=NULL){
			map.GetNextAssoc(pos,k,(void*&)p);
			ret += "<tr><td>" + CHTML::HTMLEncode(k) + "</td><td>";
			(p==NULL) ? ret += "<font color=\"red\">ERROR</font>": ret+= "<font color=\"green\">OK</font>";
			ret += "</td></tr>\r\n";
		}
		ret += "</table>\r\n</div>\r\n";
	}
	return ret;
}

CString CHTTPFirewall::CollapsibleTable(LPCTSTR Title, CString& Text, unsigned int count)
{
	//TODO: ADMIN - make collapsible table also sortable

	//title
	CString ret = "<p>";
	ret += Title;
	ret += " (" + CStringHelper::Safe_IntToAscii(count) + ")";
	if(count>0){
		ret += " <a href=\"javascript:ToggleDisplay(\'";
		ret += Title;
		ret += "\');\">+</a>";
	}
	ret += "</p>\r\n";
	//items
	ret += "<div id=\"";
	ret += Title;
	ret += "\" style=\"display:none;\">\r\n";
	ret += "<table border=\"1\" id=\"table_";
	ret += Title;
	ret += "\">\r\n";
	//ret += "<thead><tr><th>Sort</th></tr></thead>\r\n";
	ret += "<tbody>\r\n<tr><td>";
	CString contents = CHTML::HTMLEncode(Text);
	contents.Replace("|||","</td></tr>\r\n<tr><td>");
	ret += contents;
	ret += "</td></tr>\r\n</tbody>\r\n</table>\r\n</div>\r\n";

	////tabletools (uses jquery)
	//ret += " <script type=\"text/javascript\">\r\n";
	//ret += " 	  $(document).ready( function () {\r\n";
	//ret += "     $('#table_"; ret+= Title; ret+="').dataTable( {\r\n";
	//ret += "         \"sDom\": 'T<\"clear\">lfrtip',\r\n";
	//ret += "         \"oTableTools\": {\r\n";
	//ret += "             \"sSwfPath\": \"/DataTables-1.9.4/media/swf/copy_cvs_xls_pdf.swf\"\r\n";
	//ret += "         }\r\n";
	//ret += "     } );\r\n";
	//ret += " } ); \r\n";
	//ret += " </script>\r\n";

	return ret;
}

CString CHTTPFirewall::AddRow(CString Key, CString Value)
{
	return "<tr><td>" + Key + ":</td><td>" + Value + "</td></tr>\r\n";
}

CString CHTTPFirewall::GetStatisticsAdditionalCaches()
{
	return "";
}
